# TODO: repoint to dbt-artifacts when it is available
from dbt.artifacts.schemas.catalog import CatalogArtifact
import pytest

from dbt.tests.adapter.catalog import files
from dbt.tests.util import run_dbt


class CatalogRelationTypes:
    """
    Many adapters can use this test as-is. However, if your adapter contains different
    relation types or uses different strings to describe the node (e.g. 'table' instead of 'BASE TABLE'),
    then you'll need to configure this test.

    To configure this test, you'll most likely need to update either `models`
    and/or `test_relation_types_populate_correctly`. For example, `dbt-snowflake`
    supports dynamic tables and does not support materialized views. It's implementation
    might look like this:

    class TestCatalogRelationTypes:
        @pytest.fixture(scope="class", autouse=True)
        def models(self):
            yield {
                "my_table.sql": files.MY_TABLE,
                "my_view.sql": files.MY_VIEW,
                "my_dynamic_table.sql": files.MY_DYNAMIC_TABLE,
            }

        @pytest.mark.parametrize(
            "node_name,relation_type",
            [
                ("seed.test.my_seed", "BASE TABLE"),
                ("model.test.my_table", "BASE TABLE"),
                ("model.test.my_view", "VIEW"),
                ("model.test.my_dynamic_table", "DYNAMIC TABLE"),
            ],
        )
        def test_relation_types_populate_correctly(
            self, docs: CatalogArtifact, node_name: str, relation_type: str
        ):
            super().test_relation_types_populate_correctly(
                docs, node_name, relation_type
            )

    Note that we're able to configure the test case using pytest parameterization
    and call back to the original test. That way any updates to the test are incorporated
    into your adapter.
    """

    @pytest.fixture(scope="class", autouse=True)
    def seeds(self):
        return {"my_seed.csv": files.MY_SEED}

    @pytest.fixture(scope="class", autouse=True)
    def models(self):
        yield {
            "my_table.sql": files.MY_TABLE,
            "my_view.sql": files.MY_VIEW,
            "my_materialized_view.sql": files.MY_MATERIALIZED_VIEW,
        }

    @pytest.fixture(scope="class", autouse=True)
    def docs(self, project):
        run_dbt(["seed"])
        run_dbt(["run"])
        yield run_dbt(["docs", "generate"])

    @pytest.mark.parametrize(
        "node_name,relation_type",
        [
            ("seed.test.my_seed", "BASE TABLE"),
            ("model.test.my_table", "BASE TABLE"),
            ("model.test.my_view", "VIEW"),
            ("model.test.my_materialized_view", "MATERIALIZED VIEW"),
        ],
    )
    def test_relation_types_populate_correctly(
        self, docs: CatalogArtifact, node_name: str, relation_type: str
    ):
        """
        This test addresses: https://github.com/dbt-labs/dbt-core/issues/8864
        """
        assert node_name in docs.nodes
        node = docs.nodes[node_name]
        assert node.metadata.type == relation_type



# If date_spine works properly, there should be no `null` values in the resulting model

models__test_date_spine_sql = """
with generated_dates as (
    {% if target.type == 'postgres' %}
        {{ date_spine("day", "'2023-09-01'::date", "'2023-09-10'::date") }}

    {% elif target.type == 'bigquery' or target.type == 'redshift' %}
        select cast(date_day as date) as date_day
        from ({{ date_spine("day", "'2023-09-01'", "'2023-09-10'") }})

    {% else %}
        {{ date_spine("day", "'2023-09-01'", "'2023-09-10'") }}
    {% endif %}
), expected_dates as (
    {% if target.type == 'postgres' %}
        select '2023-09-01'::date as expected
        union all
        select '2023-09-02'::date as expected
        union all
        select '2023-09-03'::date as expected
        union all
        select '2023-09-04'::date as expected
        union all
        select '2023-09-05'::date as expected
        union all
        select '2023-09-06'::date as expected
        union all
        select '2023-09-07'::date as expected
        union all
        select '2023-09-08'::date as expected
        union all
        select '2023-09-09'::date as expected

    {% elif target.type == 'bigquery' or target.type == 'redshift' %}
        select cast('2023-09-01' as date) as expected
        union all
        select cast('2023-09-02' as date) as expected
        union all
        select cast('2023-09-03' as date) as expected
        union all
        select cast('2023-09-04' as date) as expected
        union all
        select cast('2023-09-05' as date) as expected
        union all
        select cast('2023-09-06' as date) as expected
        union all
        select cast('2023-09-07' as date) as expected
        union all
        select cast('2023-09-08' as date) as expected
        union all
        select cast('2023-09-09' as date) as expected

    {% else %}
        select '2023-09-01' as expected
        union all
        select '2023-09-02' as expected
        union all
        select '2023-09-03' as expected
        union all
        select '2023-09-04' as expected
        union all
        select '2023-09-05' as expected
        union all
        select '2023-09-06' as expected
        union all
        select '2023-09-07' as expected
        union all
        select '2023-09-08' as expected
        union all
        select '2023-09-09' as expected
    {% endif %}
), joined as (
    select
        generated_dates.date_day,
        expected_dates.expected
    from generated_dates
    left join expected_dates on generated_dates.date_day = expected_dates.expected
)

SELECT * from joined
"""

models__test_date_spine_yml = """
version: 2
models:
  - name: test_date_spine
    data_tests:
      - assert_equal:
          actual: date_day
          expected: expected
"""



import pytest

from dbt.tests.util import check_relations_equal, run_dbt

# Snapshot source data for the tests in this file
_seed_new_record_mode_statements = [
    """create table {database}.{schema}.seed (
    id INTEGER,
    first_name VARCHAR(50),
    last_name VARCHAR(50),
    email VARCHAR(50),
    gender VARCHAR(50),
    ip_address VARCHAR(20),
    updated_at TIMESTAMP WITHOUT TIME ZONE);""",
    """create table {database}.{schema}.snapshot_expected (
    id INTEGER,
    first_name VARCHAR(50),
    last_name VARCHAR(50),
    email VARCHAR(50),
    gender VARCHAR(50),
    ip_address VARCHAR(20),

    -- snapshotting fields
    updated_at TIMESTAMP WITHOUT TIME ZONE,
    dbt_valid_from TIMESTAMP WITHOUT TIME ZONE,
    dbt_valid_to   TIMESTAMP WITHOUT TIME ZONE,
    dbt_scd_id     TEXT,
    dbt_updated_at TIMESTAMP WITHOUT TIME ZONE,
    dbt_is_deleted TEXT
    );""",
    # seed inserts
    #  use the same email for two users to verify that duplicated check_cols values
    # are handled appropriately
    """insert into {database}.{schema}.seed (id, first_name, last_name, email, gender, ip_address, updated_at) values
(1, 'Judith', 'Kennedy', '(not provided)', 'Female', '54.60.24.128', '2015-12-24 12:19:28'),
(2, 'Arthur', 'Kelly', '(not provided)', 'Male', '62.56.24.215', '2015-10-28 16:22:15'),
(3, 'Rachel', 'Moreno', 'rmoreno2@msu.edu', 'Female', '31.222.249.23', '2016-04-05 02:05:30'),
(4, 'Ralph', 'Turner', 'rturner3@hp.com', 'Male', '157.83.76.114', '2016-08-08 00:06:51'),
(5, 'Laura', 'Gonzales', 'lgonzales4@howstuffworks.com', 'Female', '30.54.105.168', '2016-09-01 08:25:38'),
(6, 'Katherine', 'Lopez', 'klopez5@yahoo.co.jp', 'Female', '169.138.46.89', '2016-08-30 18:52:11'),
(7, 'Jeremy', 'Hamilton', 'jhamilton6@mozilla.org', 'Male', '231.189.13.133', '2016-07-17 02:09:46'),
(8, 'Heather', 'Rose', 'hrose7@goodreads.com', 'Female', '87.165.201.65', '2015-12-29 22:03:56'),
(9, 'Gregory', 'Kelly', 'gkelly8@trellian.com', 'Male', '154.209.99.7', '2016-03-24 21:18:16'),
(10, 'Rachel', 'Lopez', 'rlopez9@themeforest.net', 'Female', '237.165.82.71', '2016-08-20 15:44:49'),
(11, 'Donna', 'Welch', 'dwelcha@shutterfly.com', 'Female', '103.33.110.138', '2016-02-27 01:41:48'),
(12, 'Russell', 'Lawrence', 'rlawrenceb@qq.com', 'Male', '189.115.73.4', '2016-06-11 03:07:09'),
(13, 'Michelle', 'Montgomery', 'mmontgomeryc@scientificamerican.com', 'Female', '243.220.95.82', '2016-06-18 16:27:19'),
(14, 'Walter', 'Castillo', 'wcastillod@pagesperso-orange.fr', 'Male', '71.159.238.196', '2016-10-06 01:55:44'),
(15, 'Robin', 'Mills', 'rmillse@vkontakte.ru', 'Female', '172.190.5.50', '2016-10-31 11:41:21'),
(16, 'Raymond', 'Holmes', 'rholmesf@usgs.gov', 'Male', '148.153.166.95', '2016-10-03 08:16:38'),
(17, 'Gary', 'Bishop', 'gbishopg@plala.or.jp', 'Male', '161.108.182.13', '2016-08-29 19:35:20'),
(18, 'Anna', 'Riley', 'arileyh@nasa.gov', 'Female', '253.31.108.22', '2015-12-11 04:34:27'),
(19, 'Sarah', 'Knight', 'sknighti@foxnews.com', 'Female', '222.220.3.177', '2016-09-26 00:49:06'),
(20, 'Phyllis', 'Fox', null, 'Female', '163.191.232.95', '2016-08-21 10:35:19');""",
    # populate snapshot table
    """insert into {database}.{schema}.snapshot_expected (
    id,
    first_name,
    last_name,
    email,
    gender,
    ip_address,
    updated_at,
    dbt_valid_from,
    dbt_valid_to,
    dbt_updated_at,
    dbt_scd_id,
    dbt_is_deleted
)
select
    id,
    first_name,
    last_name,
    email,
    gender,
    ip_address,
    updated_at,
    -- fields added by snapshotting
    updated_at as dbt_valid_from,
    null::timestamp as dbt_valid_to,
    updated_at as dbt_updated_at,
    md5(id || '-' || first_name || '|' || updated_at::text) as dbt_scd_id,
    'False' as dbt_is_deleted
from {database}.{schema}.seed;""",
]

_snapshot_actual_sql = """
{% snapshot snapshot_actual %}

    {{
        config(
            unique_key='id || ' ~ "'-'" ~ ' || first_name',
        )
    }}

    select * from {{target.database}}.{{target.schema}}.seed

{% endsnapshot %}
"""

_snapshots_yml = """
snapshots:
  - name: snapshot_actual
    config:
      strategy: timestamp
      updated_at: updated_at
      hard_deletes: new_record
"""

_ref_snapshot_sql = """
select * from {{ ref('snapshot_actual') }}
"""


_invalidate_sql_statements = [
    """-- Update records 11 - 21. Change email and updated_at field.
update {schema}.seed set
    updated_at = updated_at + interval '1 hour',
    email      =  case when id = 20 then 'pfoxj@creativecommons.org' else 'new_' || email end
where id >= 10 and id <= 20;""",
    """-- Update the expected snapshot data to reflect the changes we expect to the snapshot on the next run
update {schema}.snapshot_expected set
    dbt_valid_to   = updated_at + interval '1 hour'
where id >= 10 and id <= 20;
""",
]

_update_sql = """
-- insert v2 of the 11 - 21 records

insert into {database}.{schema}.snapshot_expected (
    id,
    first_name,
    last_name,
    email,
    gender,
    ip_address,
    updated_at,
    dbt_valid_from,
    dbt_valid_to,
    dbt_updated_at,
    dbt_scd_id,
    dbt_is_deleted
)

select
    id,
    first_name,
    last_name,
    email,
    gender,
    ip_address,
    updated_at,
    -- fields added by snapshotting
    updated_at as dbt_valid_from,
    null::timestamp as dbt_valid_to,
    updated_at as dbt_updated_at,
    md5(id || '-' || first_name || '|' || updated_at::text) as dbt_scd_id,
    'False' as dbt_is_deleted
from {database}.{schema}.seed
where id >= 10 and id <= 20;
"""

# SQL to delete a record from the snapshot source data
_delete_sql = """
delete from {database}.{schema}.seed where id = 1
"""

# If the deletion worked correctly, this should return two rows, with one of them representing the deletion.
_delete_check_sql = """
select dbt_valid_to, dbt_scd_id, dbt_is_deleted from {schema}.snapshot_actual where id = 1
"""

# SQL to re-insert the deleted record from the snapshot source data
_reinsert_sql = """
insert into {database}.{schema}.seed (id, first_name, last_name, email, gender, ip_address, updated_at) values
(1, 'Judith', 'Kennedy', '(not provided)', 'Female', '54.60.24.128', '2200-01-01 12:00:00');
"""

# If the re-insertion worked correctly, this should return three rows:
#   - one for the original record
#   - one for the deletion interval
#   - one for the new record
_reinsert_check_sql = """
select dbt_valid_from, dbt_valid_to, dbt_scd_id, dbt_is_deleted from {schema}.snapshot_actual where id = 1;
"""


class BaseSnapshotNewRecordTimestampMode:
    @pytest.fixture(scope="class")
    def snapshots(self):
        return {"snapshot.sql": _snapshot_actual_sql}

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "snapshots.yml": _snapshots_yml,
            "ref_snapshot.sql": _ref_snapshot_sql,
        }

    @pytest.fixture(scope="class")
    def seed_new_record_mode_statements(self):
        return _seed_new_record_mode_statements

    @pytest.fixture(scope="class")
    def invalidate_sql_statements(self):
        return _invalidate_sql_statements

    @pytest.fixture(scope="class")
    def update_sql(self):
        return _update_sql

    @pytest.fixture(scope="class")
    def delete_sql(self):
        return _delete_sql

    @pytest.fixture(scope="class")
    def reinsert_sql(self):
        return _reinsert_sql

    @pytest.fixture(scope="class")
    def reinsert_check_sql(self):
        return _reinsert_check_sql

    def test_snapshot_new_record_mode(
        self,
        project,
        seed_new_record_mode_statements,
        invalidate_sql_statements,
        update_sql,
        delete_sql,
        reinsert_sql,
        reinsert_check_sql,
    ):
        for stmt in seed_new_record_mode_statements:
            project.run_sql(stmt)

        results = run_dbt(["snapshot"])
        assert len(results) == 1

        for stmt in invalidate_sql_statements:
            project.run_sql(stmt)

        project.run_sql(update_sql)

        results = run_dbt(["snapshot"])
        assert len(results) == 1

        check_relations_equal(project.adapter, ["snapshot_actual", "snapshot_expected"])

        project.run_sql(delete_sql)

        results = run_dbt(["snapshot"])
        assert len(results) == 1

        check_result = project.run_sql(_delete_check_sql, fetch="all")
        valid_to = 0
        scd_id = 1
        is_deleted = 2
        assert len(check_result) == 2
        assert (
            sum(
                [
                    1
                    for c in check_result
                    if c[valid_to] is None and c[scd_id] is not None and c[is_deleted] == "True"
                ]
            )
            == 1
        )
        assert (
            sum(
                [
                    1
                    for c in check_result
                    if c[valid_to] is not None
                    and c[scd_id] is not None
                    and c[is_deleted] == "False"
                ]
            )
            == 1
        )
        assert check_result[0][scd_id] != check_result[1][scd_id]

        # run snapshot with the same source data; neither insert nor update should happen
        run_dbt(["snapshot"])
        assert len(results) == 1
        check_result = project.run_sql(_delete_check_sql, fetch="all")
        assert len(check_result) == 2

        # insert the record back and run the snapshot again; update and insert expected
        project.run_sql(reinsert_sql)
        results = run_dbt(["snapshot"])
        assert len(results) == 1
        check_result = project.run_sql(reinsert_check_sql, fetch="all")
        assert len(check_result) == 3

        # delete it once again and run the snapshot; update and insert expected
        project.run_sql(delete_sql)
        results = run_dbt(["snapshot"])
        assert len(results) == 1
        check_result = project.run_sql(_delete_check_sql, fetch="all")
        assert len(check_result) == 4

        # run snapshot with the same source data; neither insert nor update should happen
        results = run_dbt(["snapshot"])
        assert len(results) == 1
        check_result = project.run_sql(_delete_check_sql, fetch="all")
        assert len(check_result) == 4



#
# Seeds
#
seeds__people = """id,first_name,last_name,email,gender,ip_address
1,Jack,Hunter,jhunter0@pbs.org,Male,59.80.20.168
2,Kathryn,Walker,kwalker1@ezinearticles.com,Female,194.121.179.35
3,Gerald,Ryan,gryan2@com.com,Male,11.3.212.243
4,Bonnie,Spencer,bspencer3@ameblo.jp,Female,216.32.196.175
5,Harold,Taylor,htaylor4@people.com.cn,Male,253.10.246.136
6,Jacqueline,Griffin,jgriffin5@t.co,Female,16.13.192.220
7,Wanda,Arnold,warnold6@google.nl,Female,232.116.150.64
8,Craig,Ortiz,cortiz7@sciencedaily.com,Male,199.126.106.13
9,Gary,Day,gday8@nih.gov,Male,35.81.68.186
10,Rose,Wright,rwright9@yahoo.co.jp,Female,236.82.178.100
"""

seeds__expected_accepted_values = """value_field,n_records
Gary,1
Rose,1
"""

seeds__expected_failing_test = """id,first_name,last_name,email,gender,ip_address
1,Jack,Hunter,jhunter0@pbs.org,Male,59.80.20.168
2,Kathryn,Walker,kwalker1@ezinearticles.com,Female,194.121.179.35
3,Gerald,Ryan,gryan2@com.com,Male,11.3.212.243
4,Bonnie,Spencer,bspencer3@ameblo.jp,Female,216.32.196.175
5,Harold,Taylor,htaylor4@people.com.cn,Male,253.10.246.136
6,Jacqueline,Griffin,jgriffin5@t.co,Female,16.13.192.220
7,Wanda,Arnold,warnold6@google.nl,Female,232.116.150.64
8,Craig,Ortiz,cortiz7@sciencedaily.com,Male,199.126.106.13
9,Gary,Day,gday8@nih.gov,Male,35.81.68.186
10,Rose,Wright,rwright9@yahoo.co.jp,Female,236.82.178.100
"""

seeds__expected_not_null_problematic_model_id = """id,first_name,last_name,email,gender,ip_address
,Gerald,Ryan,gryan2@com.com,Male,11.3.212.243
,Bonnie,Spencer,bspencer3@ameblo.jp,Female,216.32.196.175
"""

seeds__expected_unique_problematic_model_id = """unique_field,n_records
2,2
1,2
"""

#
# Schema
#
properties__schema_yml = """
version: 2

models:

  - name: fine_model
    columns:
      - name: id
        data_tests:
          - unique
          - not_null

  - name: problematic_model
    columns:
      - name: id
        data_tests:
          - unique:
              store_failures: true
          - not_null
      - name: first_name
        data_tests:
          # test truncation of really long test name
          - accepted_values:
              values:
                - Jack
                - Kathryn
                - Gerald
                - Bonnie
                - Harold
                - Jacqueline
                - Wanda
                - Craig
                # - Gary
                # - Rose

  - name: fine_model_but_with_a_no_good_very_long_name
    columns:
      - name: quite_long_column_name
        data_tests:
          # test truncation of really long test name with builtin
          - unique
"""

properties__seeds_yml = """
seeds:
  - name: people
    columns:
      - name: first_name
        tests:
          - accepted_values:
              values: ["Gary"]
              config:
                store_failures_as: table
                # 9 actual failing rows
                limit: 4
"""

#
# Models
#
models__fine_model = """
select * from {{ ref('people') }}
"""

models__file_model_but_with_a_no_good_very_long_name = """
select 1 as quite_long_column_name
"""

models__problematic_model = """
select * from {{ ref('people') }}

union all

select * from {{ ref('people') }}
where id in (1,2)

union all

select null as id, first_name, last_name, email, gender, ip_address from {{ ref('people') }}
where id in (3,4)
"""

#
# Tests
#
tests__failing_test = """
select * from {{ ref('fine_model') }}
"""

tests__passing_test = """
select * from {{ ref('fine_model') }}
where false
"""



import pytest

from dbt.tests.adapter.utils import base_utils, fixture_date


class BaseDate(base_utils.BaseUtils):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_date.yml": fixture_date.models__test_date_yml,
            "test_date.sql": self.interpolate_macro_namespace(
                fixture_date.models__test_date_sql, "date"
            ),
        }


class TestDate(BaseDate):
    pass



from typing import Type
from unittest.mock import MagicMock

from dbt_common.exceptions import DbtRuntimeError
import pytest

from dbt.adapters.base.impl import BaseAdapter


class BaseCalculateFreshnessMethod:
    """Tests the behavior of the calculate_freshness_from_customsql method for the relevant adapters.

    The base method is meant to throw the appropriate custom exception when calculate_freshness_from_customsql
    fails.
    """

    @pytest.fixture(scope="class")
    def valid_sql(self) -> str:
        """Returns a valid statement for issuing as a validate_sql query.

        Ideally this would be checkable for non-execution. For example, we could use a
        CREATE TABLE statement with an assertion that no table was created. However,
        for most adapter types this is unnecessary - the EXPLAIN keyword has exactly the
        behavior we want, and here we are essentially testing to make sure it is
        supported. As such, we return a simple SELECT query, and leave it to
        engine-specific test overrides to specify more detailed behavior as appropriate.
        """

        return "select now()"

    @pytest.fixture(scope="class")
    def invalid_sql(self) -> str:
        """Returns an invalid statement for issuing a bad validate_sql query."""

        return "Let's run some invalid SQL and see if we get an error!"

    @pytest.fixture(scope="class")
    def expected_exception(self) -> Type[Exception]:
        """Returns the Exception type thrown by a failed query.

        Defaults to dbt_common.exceptions.DbtRuntimeError because that is the most common
        base exception for adapters to throw."""
        return DbtRuntimeError

    @pytest.fixture(scope="class")
    def mock_relation(self):
        mock = MagicMock()
        mock.__str__ = lambda x: "test.table"
        return mock

    def test_calculate_freshness_from_custom_sql_success(
        self, adapter: BaseAdapter, valid_sql: str, mock_relation
    ) -> None:
        with adapter.connection_named("test_freshness_custom_sql"):
            adapter.calculate_freshness_from_custom_sql(mock_relation, valid_sql)

    def test_calculate_freshness_from_custom_sql_failure(
        self,
        adapter: BaseAdapter,
        invalid_sql: str,
        expected_exception: Type[Exception],
        mock_relation,
    ) -> None:
        with pytest.raises(expected_exception=expected_exception):
            with adapter.connection_named("test_infreshness_custom_sql"):
                adapter.calculate_freshness_from_custom_sql(mock_relation, invalid_sql)


class TestCalculateFreshnessMethod(BaseCalculateFreshnessMethod):
    pass



import pytest

from dbt.tests.fixtures.project import write_project_files
from dbt.tests.util import check_relations_equal, run_dbt


tests__get_columns_in_relation_sql = """
{% set columns = adapter.get_columns_in_relation(ref('model')) %}
{% set limit_query = 0 %}
{% if (columns | length) == 0 %}
    {% set limit_query = 1 %}
{% endif %}

select 1 as id limit {{ limit_query }}

"""

models__upstream_sql = """
select 1 as id

"""

models__expected_sql = """
-- make sure this runs after 'model'
-- {{ ref('model') }}
select 2 as id

"""

models__model_sql = """

{% set upstream = ref('upstream') %}

{% if execute %}
    {# don't ever do any of this #}
    {%- do adapter.drop_schema(upstream) -%}
    {% set existing = adapter.get_relation(upstream.database, upstream.schema, upstream.identifier) %}
    {% if existing is not none %}
        {% do exceptions.raise_compiler_error('expected ' ~ ' to not exist, but it did') %}
    {% endif %}

    {%- do adapter.create_schema(upstream) -%}

    {% set sql = create_view_as(upstream, 'select 2 as id') %}
    {% do run_query(sql) %}
{% endif %}


select * from {{ upstream }}

"""


class BaseAdapterMethod:
    """
    This test will leverage the following adapter methods:
        get_relation
        get_columns_in_relation
        drop_schema
        create_schema
    It will aims to make sure drop_shema actually works, for more context
    checkout #1983
    """

    @pytest.fixture(scope="class")
    def tests(self):
        return {"get_columns_in_relation.sql": tests__get_columns_in_relation_sql}

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "upstream.sql": models__upstream_sql,
            "expected.sql": models__expected_sql,
            "model.sql": models__model_sql,
        }

    @pytest.fixture(scope="class")
    def project_files(
        self,
        project_root,
        tests,
        models,
    ):
        write_project_files(project_root, "tests", tests)
        write_project_files(project_root, "models", models)

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "name": "adapter_methods",
        }

    @pytest.fixture(autouse=True)
    def clean_up(self, project):
        yield
        with project.adapter.connection_named("__test"):
            relation = project.adapter.Relation.create(
                database=project.database, schema=project.test_schema
            )
            project.adapter.drop_schema(relation)

    pass

    # snowflake need all tables in CAP name
    @pytest.fixture(scope="class")
    def equal_tables(self):
        return ["model", "expected"]

    def test_adapter_methods(self, project, equal_tables):
        run_dbt(["compile"])  # trigger any compile-time issues
        result = run_dbt()
        assert len(result) == 3
        check_relations_equal(project.adapter, equal_tables)


class TestBaseCaching(BaseAdapterMethod):
    pass



# safe_cast

seeds__data_safe_cast_csv = """field,output
abc,abc
123,123
,
"""


models__test_safe_cast_sql = """
with data as (

    select * from {{ ref('data_safe_cast') }}

)

select
    {{ safe_cast('field', api.Column.translate_type('string')) }} as actual,
    output as expected

from data
"""


models__test_safe_cast_yml = """
version: 2
models:
  - name: test_safe_cast
    data_tests:
      - assert_equal:
          actual: actual
          expected: expected
"""



import pytest

from dbt.tests.adapter.basic import files
from dbt.tests.util import (
    check_relation_types,
    check_relations_equal,
    check_result_nodes_by_name,
    relation_from_name,
    run_dbt,
)


class BaseSimpleMaterializations:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "view_model.sql": files.base_view_sql,
            "table_model.sql": files.base_table_sql,
            "swappable.sql": files.base_materialized_var_sql,
            "schema.yml": files.schema_base_yml,
        }

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "base.csv": files.seeds_base_csv,
        }

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "name": "base",
        }

    @pytest.fixture(autouse=True)
    def clean_up(self, project):
        yield
        with project.adapter.connection_named("__test"):
            relation = project.adapter.Relation.create(
                database=project.database, schema=project.test_schema
            )
            project.adapter.drop_schema(relation)

    pass

    def test_base(self, project):
        # seed command
        results = run_dbt(["seed"])
        # seed result length
        assert len(results) == 1

        # run command
        results = run_dbt()
        # run result length
        assert len(results) == 3

        # names exist in result nodes
        check_result_nodes_by_name(results, ["view_model", "table_model", "swappable"])

        # check relation types
        expected = {
            "base": "table",
            "view_model": "view",
            "table_model": "table",
            "swappable": "table",
        }
        check_relation_types(project.adapter, expected)

        # base table rowcount
        relation = relation_from_name(project.adapter, "base")
        result = project.run_sql(f"select count(*) as num_rows from {relation}", fetch="one")
        assert result[0] == 10

        # relations_equal
        check_relations_equal(project.adapter, ["base", "view_model", "table_model", "swappable"])

        # check relations in catalog
        catalog = run_dbt(["docs", "generate"])
        assert len(catalog.nodes) == 4
        assert len(catalog.sources) == 1

        # run_dbt changing materialized_var to view
        if project.test_config.get("require_full_refresh", False):  # required for BigQuery
            results = run_dbt(
                ["run", "--full-refresh", "-m", "swappable", "--vars", "materialized_var: view"]
            )
        else:
            results = run_dbt(["run", "-m", "swappable", "--vars", "materialized_var: view"])
        assert len(results) == 1

        # check relation types, swappable is view
        expected = {
            "base": "table",
            "view_model": "view",
            "table_model": "table",
            "swappable": "view",
        }
        check_relation_types(project.adapter, expected)

        # run_dbt changing materialized_var to incremental
        results = run_dbt(["run", "-m", "swappable", "--vars", "materialized_var: incremental"])
        assert len(results) == 1

        # check relation types, swappable is table
        expected = {
            "base": "table",
            "view_model": "view",
            "table_model": "table",
            "swappable": "table",
        }
        check_relation_types(project.adapter, expected)


class TestSimpleMaterializations(BaseSimpleMaterializations):
    pass



# split_part

seeds__data_split_part_csv = """parts,split_on,result_1,result_2,result_3,result_4
a|b|c,|,a,b,c,c
1|2|3,|,1,2,3,3
,|,,,,
"""


models__test_split_part_sql = """
with data as (

    select * from {{ ref('data_split_part') }}

)

select
    {{ split_part('parts', 'split_on', 1) }} as actual,
    result_1 as expected

from data

union all

select
    {{ split_part('parts', 'split_on', 2) }} as actual,
    result_2 as expected

from data

union all

select
    {{ split_part('parts', 'split_on', 3) }} as actual,
    result_3 as expected

from data

union all

select
    {{ split_part('parts', 'split_on', -1) }} as actual,
    result_4 as expected

from data
"""


models__test_split_part_yml = """
version: 2
models:
  - name: test_split_part
    data_tests:
      - assert_equal:
          actual: actual
          expected: expected
"""



# string_literal

models__test_string_literal_sql = """
select {{ string_literal("abc") }} as actual, 'abc' as expected union all
select {{ string_literal("1") }} as actual, '1' as expected union all
select {{ string_literal("") }} as actual, '' as expected union all
select {{ string_literal(none) }} as actual, 'None' as expected
"""


models__test_string_literal_yml = """
version: 2
models:
  - name: test_string_literal
    data_tests:
      - assert_equal:
          actual: actual
          expected: expected
"""



import pytest

from dbt.tests.adapter.store_test_failures_tests import fixtures
from dbt.tests.util import check_relations_equal, run_dbt


# used to rename test audit schema to help test schema meet max char limit
# the default is _dbt_test__audit but this runs over the postgres 63 schema name char limit
# without which idempotency conditions will not hold (i.e. dbt can't drop the schema properly)
TEST_AUDIT_SCHEMA_SUFFIX = "dbt_test__aud"


class StoreTestFailuresBase:
    @pytest.fixture(scope="function", autouse=True)
    def setUp(self, project):
        self.test_audit_schema = f"{project.test_schema}_{TEST_AUDIT_SCHEMA_SUFFIX}"
        run_dbt(["seed"])
        run_dbt(["run"])

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "people.csv": fixtures.seeds__people,
            "expected_accepted_values.csv": fixtures.seeds__expected_accepted_values,
            "expected_failing_test.csv": fixtures.seeds__expected_failing_test,
            "expected_not_null_problematic_model_id.csv": fixtures.seeds__expected_not_null_problematic_model_id,
            "expected_unique_problematic_model_id.csv": fixtures.seeds__expected_unique_problematic_model_id,
        }

    @pytest.fixture(scope="class")
    def tests(self):
        return {
            "failing_test.sql": fixtures.tests__failing_test,
            "passing_test.sql": fixtures.tests__passing_test,
        }

    @pytest.fixture(scope="class")
    def properties(self):
        return {"schema.yml": fixtures.properties__schema_yml}

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "fine_model.sql": fixtures.models__fine_model,
            "fine_model_but_with_a_no_good_very_long_name.sql": fixtures.models__file_model_but_with_a_no_good_very_long_name,
            "problematic_model.sql": fixtures.models__problematic_model,
        }

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "seeds": {
                "quote_columns": False,
                "test": self.column_type_overrides(),
            },
            "data_tests": {"+schema": TEST_AUDIT_SCHEMA_SUFFIX},
        }

    def column_type_overrides(self):
        return {}

    def run_tests_store_one_failure(self, project):
        run_dbt(["test"], expect_pass=False)

        # one test is configured with store_failures: true, make sure it worked
        check_relations_equal(
            project.adapter,
            [
                f"{self.test_audit_schema}.unique_problematic_model_id",
                "expected_unique_problematic_model_id",
            ],
        )

    def run_tests_store_failures_and_assert(self, project):
        # make sure this works idempotently for all tests
        run_dbt(["test", "--store-failures"], expect_pass=False)
        results = run_dbt(["test", "--store-failures"], expect_pass=False)

        # compare test results
        actual = [(r.status, r.failures) for r in results]
        expected = [
            ("pass", 0),
            ("pass", 0),
            ("pass", 0),
            ("pass", 0),
            ("fail", 2),
            ("fail", 2),
            ("fail", 2),
            ("fail", 10),
        ]
        assert sorted(actual) == sorted(expected)

        # compare test results stored in database
        check_relations_equal(
            project.adapter, [f"{self.test_audit_schema}.failing_test", "expected_failing_test"]
        )
        check_relations_equal(
            project.adapter,
            [
                f"{self.test_audit_schema}.not_null_problematic_model_id",
                "expected_not_null_problematic_model_id",
            ],
        )
        check_relations_equal(
            project.adapter,
            [
                f"{self.test_audit_schema}.unique_problematic_model_id",
                "expected_unique_problematic_model_id",
            ],
        )
        check_relations_equal(
            project.adapter,
            [
                f"{self.test_audit_schema}.accepted_values_problemat"
                "ic_mo_c533ab4ca65c1a9dbf14f79ded49b628",
                "expected_accepted_values",
            ],
        )


class BaseStoreTestFailures(StoreTestFailuresBase):
    @pytest.fixture(scope="function")
    def clean_up(self, project):
        yield
        with project.adapter.connection_named("__test"):
            relation = project.adapter.Relation.create(
                database=project.database, schema=self.test_audit_schema
            )
            project.adapter.drop_schema(relation)

            relation = project.adapter.Relation.create(
                database=project.database, schema=project.test_schema
            )
            project.adapter.drop_schema(relation)

    def column_type_overrides(self):
        return {
            "expected_unique_problematic_model_id": {
                "+column_types": {
                    "n_records": "bigint",
                },
            },
            "expected_accepted_values": {
                "+column_types": {
                    "n_records": "bigint",
                },
            },
        }

    def test__store_and_assert(self, project, clean_up):
        self.run_tests_store_one_failure(project)
        self.run_tests_store_failures_and_assert(project)


class BaseStoreTestFailuresLimit(BaseStoreTestFailures):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "people.csv": fixtures.seeds__people,
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {}

    @pytest.fixture(scope="class")
    def tests(self):
        return {}

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {}

    @pytest.fixture(scope="class")
    def properties(self):
        return {"_seeds.yml": fixtures.properties__seeds_yml}

    def test__store_and_assert(self, project, clean_up):
        pass

    def test_store_limit(self, project, clean_up):
        results = run_dbt(["test"], expect_pass=False)
        # there are 9 actual failing rows, but the test `limit` config has a value of 4
        assert results.results[0].failures == 4
        relation_name = results.results[0].node.relation_name
        sql_result = project.run_sql(f"select count(*) as cnt from {relation_name}", fetch="one")
        count = sql_result[0] if sql_result is not None else None
        # make sure the table also only has 4 rows (not 9!)
        assert count == 4, f"The test failure count {count} doesn't match the config `limit` of 4"


class TestStoreTestFailures(BaseStoreTestFailures):
    pass


class TestStoreTestFailuresLimit(BaseStoreTestFailuresLimit):
    pass
